#ifndef _COURS_ARRAY_H
#define _COURS_ARRAY_H
// ----------------------------------------------------------------------------
//
// a valarray with smart pointers that should work for
//
//	array<T,vector<T> >	=> T = GiNaC::ex
//	array<T,valarray<T> >	=> T = double 
//
// because valarray<GiNaC::ex> segfaults in expressions, e.g. x-y
// 
// Location in memory are contiguous, unstead when using
//	M = slice_array<T> and such...
// TODO: set a contiguous_tag to use iterators or indexes ?
//
// ----------------------------------------------------------------------------
#include "smart_pointer.h"
#include <vector>

template <typename T, typename M = std::vector<T> >
class array : public smart_pointer<M> {
    public:
	typedef T value_type;
	typedef M data_type;
	typedef smart_pointer<M> pointer_type;
	typedef size_t size_type;
	array (size_type = 0, const T& = 0);
	size_type size () const;
	void resize (size_type);
	const T& operator[] (size_type) const;
	T& operator[] (size_type);

	array<T,M>& operator+= (const array<T,M>&);
	array<T,M>& operator-= (const array<T,M>&);

    // internals
    template <typename BinaryOperator>
    static void array_binop (const M& u, const M& v, BinaryOperator f, M& w);
};
// ============================================================================
// cstors, accessors
// ============================================================================
template <typename T, typename M>
array<T,M>::array (size_type n, const T& initval)
  : smart_pointer<M> (new_macro(M(n)))
{
	std::fill_n (&(pointer_type::data()[0]), n, initval);
}
template <typename T, typename M>
typename array<T,M>::size_type
array<T,M>::size () const
{
	return pointer_type::data().size();
}
template <typename T, typename M>
void
array<T,M>::resize (size_type n)
{
	return pointer_type::data().resize(n);
}
template <typename T, typename M>
const T&
array<T,M>::operator[] (size_type i) const
{
	return pointer_type::data().operator[] (i);
}
template <typename T, typename M>
T&
array<T,M>::operator[] (size_type i)
{
	return pointer_type::data().operator[] (i);
}
// ============================================================================
// algebra
// ============================================================================
// array: u+v, ...
// ----------------------------------------------------------------------------
template <typename T, typename M> 
template <typename BinaryOperator> 
void 
array<T,M>::array_binop (const M& u, const M& v, BinaryOperator f, M& w)
{
	for (typename array<T,M>::size_type i = 0; i < w.size(); i++)
	    w [i] = f (u [i], v[i]);
}
#define array_binary_operator(op, name)					\
template <typename T, typename M>						\
array<T,M>								\
operator op (const array<T,M>& u, const array<T,M>& v)			\
{									\
	check_macro (u.size() == v.size(), 				\
		"incompatible sizes for binary operator");		\
	array<T,M> w (u.size());					\
	array<T,M>::array_binop (u.data(), v.data(), std::name<T>(),	\
				 w.data());				\
	return w;							\
}
       array_binary_operator (+, plus)
       array_binary_operator (-, minus)
#undef array_binary_operator

// ----------------------------------------------------------------------------
// hybrid: c*u, ...
// ----------------------------------------------------------------------------
template <typename T, typename M, typename BinaryOperator>
array<T,M>
array_binop (const T& c, const array<T,M>& u, BinaryOperator f)
{
	array<T,M> w (u.size());
	for (typename array<T,M>::size_type i = 0; i < w.size(); i++) 
	    w [i] = f (c, u [i]);
	return w;
}
template <typename T, typename M>
array<T,M>
operator* (const T& c, const array<T,M>& u)
{
	return array_binop (c, u, std::multiplies<T>());
}
// ----------------------------------------------------------------------------
// computed assignmen: u += v, ...
// ----------------------------------------------------------------------------
#define computed_assignment_def(op)					\
template <typename T, typename M>						\
array<T,M>&								\
array<T,M>::operator op (const array<T,M>& v)				\
{									\
	check_macro (size() == v.size(),				\
		"incompatible sizes for computed assignment");		\
	for (size_type i = 0; i < size(); i++)				\
	    operator[] (i) op v [i];					\
	return *this;							\
}
	computed_assignment_def (+=)
	computed_assignment_def (-=)
#undef  computed_assignment_def

#ifdef TODO
#endif // TODO

#endif // _COURS_ARRAY_H
