#ifndef _COURS_JACOBI_GAUSS_LOBATTO_DEFLATION_H
#define _COURS_JACOBI_GAUSS_LOBATTO_DEFLATION_H
//
// compute the Gauss-Lobatto collocation points for Jacobi (a,b) polynomial at order N
//
//  x [0:N-1] : the colocation points in decreasing order from +1 to -1
//  w [0:N-1] : the associated weights
//
// by polynom deflation.
//
// reference : C. Canuto, M. Y. Hussaini, A. Quarteroni and T. A. Zang
// Spectral methods in Fluid Dynamics
// Springer Verlag, 1988
// page 525
//
#include "jacobi.h"
#include <math.h>
#include "jacobi.h"
#include "numeric_constant.h"

template <typename Size, typename T, typename Container>
void
jacobi_gauss_lobatto_deflation (const T& a, const T& b, Size N, Container& x) // , Container& w)
{
	T    epsilon = 1000*numeric_constant<T>::epsilon();
	Size k_stop  =      numeric_constant<T>::digits10();

	const T pi = numeric_constant<T>::pi();
	T theta = 0.5*pi/(2*N + 1);
	T cd = cos (2*theta);
	T sd = sin (2*theta);
	T cs = cos (theta);
	T ss = sin (theta);

	T pnp1p, pdnp1p, pnp, pdnp, pnm1p, pdnm1p; // at x = +1 (1p)
	T pnp1m, pdnp1m, pnm, pdnm, pnm1m, pdnm1m; // at x = -1 (1m)
	jacobi (a, b, N+1, +1., pnp1p, pdnp1p, pnp, pdnp, pnm1p, pdnm1p);
	jacobi (a, b, N+1, -1., pnp1m, pdnp1m, pnm, pdnm, pnm1m, pdnm1m);
	T det = pnp*pnm1m - pnm*pnm1p;
	T aa  = - (pnp1p*pnm1m - pnp1m*pnm1p)/det;
	T bb  = - (pnp1m*pnp   - pnp1p*pnm)  /det;

	// compute the first half of the root by polynomial deflation
	x [0] = 1;
	for (Size j = 2; j <= (N+1)/2; j++) {

	    T xj = cs;
	    T delta_xj = 1;
	    for (Size k = 1; k <= k_stop && fabs(delta_xj) >= epsilon; k++) {
	      T pnp1, pdnp1, pn, pdn, pnm1, pdnm1;
	      jacobi (a, b, N+1, xj, pnp1, pdnp1, pn, pdn, pnm1, pdnm1);
	      T p  = pnp1  + aa*pn  + bb*pnm1;
	      T dp = pdnp1 + aa*pdn + bb*pdnm1;
	      T sum = 0;
	      for (Size i = 1; i <= j-1; i++)
		sum += 1/(xj - x [i-1]);
	      delta_xj = - p/(dp - sum*p);
	      xj += delta_xj;
	    }
	    if (fabs(delta_xj) >= epsilon) {
		    std::cerr << "x ["<<j-1<<"]: error = " << delta_xj << " at step " << k_stop << std::endl;
	    }
	    x [j-1] = xj;
	    T cs_save = cs*cd - ss*sd;
	    ss        = cs*sd + ss*cd;
	    cs = cs_save;
	}
	// use symmetry for second half of the root
	for (Size i = 0; i < (N+1)/2; i++)
	    x [N-i] = - x [i];
	if (N % 2 == 0) x [N/2] = 0;
}
#endif // _COURS_JACOBI_GAUSS_LOBATTO_DEFLATION_H
