#include "compiler.h"
#include <cmath>
#include <vector>
// -----------------------------------------------------------------------------------
// TODO : with some smart_pointer class
// -----------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------
// node
// -----------------------------------------------------------------------------------
template <typename T>
class node {
    public:
	node(const T& x = T(0)) : m_v(x), g(), m_rc(1) {}
	T& value() { return m_v; }
	const T& value() const { return m_v; }
	T& deriv(const int idx) { return g[idx]; }
	const T& deriv(const int idx) const { return g[idx]; }
	void resize_grad(size_t n) {
            g.resize(n);
            std::fill (g.begin(), g.end(), T(0));
	}
	size_t size() const { return g.size(); }
	int rc() const { return m_rc; }
	int inc_rc() { return ++m_rc; }
	int dec_rc() { return --m_rc; }
	void pv(const node& v);
	void mv(const node& v);
	void pav(const T& a, const node& v);
	void mav(const T& a, const node& v);
    protected:
	T m_v;
	std::vector<T> g;
	int m_rc;
};
template <class T>
void node<T>::pv(const node<T>& v) {
	if (v.g.size() == 0) return;
	if (g.size() == 0) {
	    g.resize(v.size());
	    for(size_t i=0; i < g.size(); ++i) g[i] = v.g[i];
	} else {
	    assert_macro(g.size() == v.g.size(), ".diff(i,n) called with two different values of"
		" n in same expression; "<< g.size() << " and " << v.g.size());
	    for(size_t i=0; i < g.size(); ++i) g[i] += v.g[i];
	}
}
template <class T>
void node<T>::mv(const node<T>& v) {
	if (v.g.size() == 0) return;
	if (g.size() == 0) {
	    g.resize(v.size());
	    for(size_t i=0; i < g.size(); ++i) g[i] = -v.g[i];
	} else {
	    assert_macro(g.size() == v.g.size(), ".diff(i,n) called with two different values of"
		" n in same expression; "<< g.size() << " and " << v.g.size());
	    for(size_t i=0; i < g.size(); ++i) g[i] -= v.g[i];
	}
}
template <class T>
void node<T>::pav(const T& a, const node<T>& v) {
	if (v.g.size() == 0) return;
	if (g.size() == 0) {
	    g.resize(v.size());
	    for(size_t i=0; i < g.size(); ++i) g[i] = a * v.g[i];
	} else {
	    assert_macro(g.size() == v.g.size(), ".diff(i,n) called with two different values of"
		" n in same expression; "<< g.size() << " and " << v.g.size());
	    for(size_t i=0; i < g.size(); ++i) g[i] += a * v.g[i];
	}
}
template <class T>
void node<T>::mav(const T& a, const node<T>& v) {
	if (v.g.size() == 0) return;
	if (g.size() == 0) {
	    g.resize(v.size());
	    for(size_t i=0; i < g.size(); ++i) g[i] = - a * v.g[i];
	} else {
	    assert_macro(g.size() == v.g.size(), ".diff(i,n) called with two different values of"
		" n in same expression; "<< g.size() << " and " << v.g.size());
	    for(size_t i=0; i < g.size(); ++i) g[i] -= a * v.g[i];
	}
}
// -----------------------------------------------------------------------------------
// op
// -----------------------------------------------------------------------------------
// This class is the baseclass of all operation met during the forward sweep. 
// It contains a pointer to a node<T> where the result of the operation is stored, val.
// -----------------------------------------------------------------------------------
template <class T>
class op {
    public:
	virtual ~op() { dec_rc(); }
	void inc_rc() const {
	    basic_assert_macro(m_val);
	    m_val->inc_rc();
        }
	virtual void dec_rc() {
	    if (m_val && m_val->dec_rc()==0) {
		delete_macro(m_val); m_val=0;
	    }
        }
	/* Dummy virtual functions */
	virtual op<T>* copy() const { return 0; }
	virtual void propagate() {}
	virtual void propagateop() {}
    protected:
	op() : m_val(0) {}
	op(const T &x) : m_val(new_copy_macro(node<T>,x)) {}
	op(const op<T> &x) : m_val(x.m_val) { if (m_val) inc_rc(); }
	template <class U> friend class dag;
        template <class U> friend class add_op;
        template <class U> friend class mult_op;
        template <class U> friend class sqrt_op;
        template <class U> friend class sin_op;
    // data:
	node<T> *m_val;
};
// -----------------------------------------------------------------------------------
// binary_op
// -----------------------------------------------------------------------------------
// This class represents binary operators. It has two branches, one for each
// of the expresions (operators) it operates on. The pointer o1 refers to
// the left operand and o2 to the right operand.
// -----------------------------------------------------------------------------------
template <class T>
class binary_op: public op<T> { 
    public:
	virtual ~binary_op() {
	    if (op<T>::m_val != 0 && op<T>::m_val->rc()==1) {
		if (m_o1 != 0) {m_o1->propagate(); delete_macro(m_o1); m_o1=0;}
		if (m_o2 != 0) {m_o2->propagate(); delete_macro(m_o2); m_o2=0;}
	    }
        }
	virtual void propagate() {
	    if (op<T>::m_val != 0 && op<T>::m_val->rc()==1) {
		this->propagateop();
		if (m_o1 != 0) {m_o1->propagate(); delete_macro(m_o1); m_o1=0;}
		if (m_o2 != 0) {m_o2->propagate(); delete_macro(m_o2); m_o2=0;}
	    }
	    this->dec_rc();
        }
    protected:
	binary_op()                               : op<T>(),  m_o1(0), m_o2(0) {}
	binary_op(const T& x)                     : op<T>(x), m_o1(0), m_o2(0) {}
	binary_op(const T& x, op<T>* a, op<T>* b) : op<T>(x), m_o1(a), m_o2(b) {}
	binary_op(const binary_op<T>& x)          : op<T>(x), m_o1(x.m_o1), m_o2(x.m_o2) {}
    // data:
	op<T> *m_o1;
	op<T> *m_o2;
};
// -----------------------------------------------------------------------------------
// unary_op
// -----------------------------------------------------------------------------------
// This class represents unary operators. It has one branch for the
// expresion (operator) it operates on. The pointer o1 refers to this.
// -----------------------------------------------------------------------------------
template <class T>
class unary_op: public op<T> {
    public:
	virtual ~unary_op() {
	    if (op<T>::m_val != 0 && op<T>::m_val->rc() == 1) {
		if (m_o1 != 0) {
		    m_o1->propagate();
		    delete_macro(m_o1); m_o1=0;
		}
	    }
        }
	virtual void propagate() {
	    if (op<T>::m_val != 0) {
		basic_assert_macro(op<T>::m_val->rc()==1);
		this->propagateop();
		if (m_o1 != 0) {
		    m_o1->propagate();
		    delete_macro(m_o1); m_o1=0;
		}
	    }
	    this->dec_rc();
        }
    protected:
	unary_op()                     : op<T>(),  m_o1(0) {}
	unary_op(const T& x)           : op<T>(x), m_o1(0) {}
	unary_op(const T& x, op<T>* a) : op<T>(x), m_o1(a) {}
	unary_op(const unary_op<T>& x) : op<T>(x), m_o1(x.m_o1) {}
    // data:
	op<T> *m_o1;
};
// -----------------------------------------------------------------------------------
// dag
// -----------------------------------------------------------------------------------
// This class is used as a substitute for variables. No (sub-)expression
// can be referenced directly more than once unless it is done by objects
// of this class. This class does often not contain its "own" node<T> but
// just refers to a node<T> from an underlying operator.
// -----------------------------------------------------------------------------------
template <class T>
class dag : public unary_op<T> { 
    public:
	dag()                : unary_op<T>() {}
	dag(const T &x)      : unary_op<T>(x) {}
	dag(const dag<T> &x) : unary_op<T>(x) {}
	dag(const op<T> &x)  : unary_op<T>() {
	    unary_op<T>::m_val = x.m_val;
	    unary_op<T>::m_o1  = x.copy();
	    this->inc_rc();
        }
	virtual ~dag() {
	    if (unary_op<T>::m_val != 0 && unary_op<T>::m_val->rc() == 2 && unary_op<T>::m_o1 != 0)
		this->dec_rc(); 
        }
	virtual void dec_rc() {
	    if (unary_op<T>::m_o1 != 0) {
	        basic_assert_macro(unary_op<T>::m_val != 0);
	        if (unary_op<T>::m_val->dec_rc() == 1) {
		    delete_macro(unary_op<T>::m_o1);
		    unary_op<T>::m_o1  = 0;
		    unary_op<T>::m_val = 0; // Since deleted from o1.
	        }
	    } else { 
	        unary_op<T>::dec_rc();
	    }
        }
	virtual void propagate() {
	    basic_assert_macro(unary_op<T>::m_val != 0 || unary_op<T>::m_o1 == 0);
	    if (unary_op<T>::m_o1 != 0 && unary_op<T>::m_val->rc() == 2) {
	        unary_op<T>::dec_rc();
	        unary_op<T>::m_o1->propagate();
	        delete_macro(unary_op<T>::m_o1);
	    } else {
	        unary_op<T>::dec_rc();
	    }
	    unary_op<T>::m_o1  = 0;
	    unary_op<T>::m_val = 0;
        }
	virtual op<T>* copy() const {
	    return new_copy_macro(dag<T>,*this);
	}
	void diff(size_t idx, size_t n) {
	    basic_assert_macro(unary_op<T>::m_val != 0);
	    unary_op<T>::m_val->resize_grad(n);
	    unary_op<T>::m_val->deriv(idx) = T(1);
	    dag<T> tmp (unary_op<T>::m_val->value());
	    this->propagate();
	    (*this) = tmp ;
	    unary_op<T>::m_o1 = 0;
        }
	T x() {
	    if (unary_op<T>::m_val == 0) return T(0);
	    return unary_op<T>::m_val->value();
	}
	T d(size_t i) {
	    if (unary_op<T>::m_val == 0 || unary_op<T>::m_val->size() == 0) return T(0);
	    basic_assert_macro(i < unary_op<T>::m_val->size());
	    return unary_op<T>::m_val->deriv(i);
        }
  // This is all overloadings of the = operator
	dag<T>& operator = (const T& x) {
	    dec_rc();
	    unary_op<T>::m_val = new_copy_macro(node<T>,x);
	    unary_op<T>::m_o1  = 0;
	    return *this;
        }
	dag<T>& operator = (const dag<T>& x) {
	    dec_rc();
	    unary_op<T>::m_val = x.m_val;
	    if (unary_op<T>::m_val != 0) this->inc_rc();
	    unary_op<T>::m_o1  = x.m_o1;
	    return *this;
        }
	dag<T>& operator = (const op<T>& x) {
	    this->dec_rc();
	    unary_op<T>::m_val = x.m_val;
	    this->inc_rc();
	    unary_op<T>::m_o1  = x.copy();
	    return *this;
        }
  // Declaration of overloading of (?)=
	dag<T>& operator += (const op<T>& x);
	dag<T>& operator -= (const op<T>& x);
	dag<T>& operator *= (const op<T>& x);
	dag<T>& operator /= (const op<T>& x);
	dag<T>& operator += (const dag<T>& x);
	dag<T>& operator -= (const dag<T>& x);
	dag<T>& operator *= (const dag<T>& x);
	dag<T>& operator /= (const dag<T>& x);
	dag<T>& operator += (const T& x);
	dag<T>& operator -= (const T& x);
	dag<T>& operator *= (const T& x);
	dag<T>& operator /= (const T& x);
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION ADDITION FORMULAS               */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* --------------------------- op + op --------------------------------- */

template <class T>
class add_op: public binary_op<T> {
    public:
	add_op(const add_op<T>& o);
	add_op(const op<T>& a, const op<T>& b);
	virtual ~add_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*            BACKWARD DIFFERENTIATION SUBTRACTION FORMULAS              */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* --------------------------- op - op --------------------------------- */

template <class T>
class sub_op: public binary_op<T> {
    public:
	sub_op(const sub_op&);
	sub_op(const op<T>& a, const op<T>& b);
	virtual ~sub_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*          BACKWARD DIFFERENTIATION MULTIPLICATION FORMULAS             */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* --------------------------- op * op --------------------------------- */

template <class T>
class mult_op: public binary_op<T> {
    public:
	mult_op(const mult_op<T>&);
	mult_op(const op<T>& a, const op<T>& b);
	virtual ~mult_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION DIVISION FORMULAS               */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* --------------------------- op / op --------------------------------- */

template <class T>
class div_op: public binary_op<T> {
    public:
	div_op(const div_op&);
	div_op(const op<T>& a, const op<T>& b);
	virtual ~div_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*                BACKWARD DIFFERENTIATION POWER FORMULAS                */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* -------------------------- pow(op,op) ------------------------------- */

template <class T>
class pow_op: public binary_op<T> {
    public:
	pow_op(const pow_op&);
	pow_op(const op<T>& a, const op<T>& b);
	virtual ~pow_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*            BACKWARD DIFFERENTIATION UNARY + FORMULA                   */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class uadd_op: public unary_op<T> {
    public:
	uadd_op(const uadd_op&);
	uadd_op(const op<T>& a);
	virtual ~uadd_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*            BACKWARD DIFFERENTIATION UNARY - FORMULA                   */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class usub_op: public unary_op<T> {
    public:
	usub_op(const usub_op&);
	usub_op(const op<T>& a);
	virtual ~usub_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*          BACKWARD DIFFERENTIATION pow(op,int) FORMULA                 */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class pow_n_op: public unary_op<T> {
    public:
	pow_n_op(const pow_n_op&);
	pow_n_op(const op<T>& a, int ib);
	virtual ~pow_n_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
    protected:
	int m_b;
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION sqr FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */
template <class T>
class sqr_op: public unary_op<T> {
    public:
	sqr_op(const sqr_op&);
	sqr_op(const op<T>& a);
	virtual ~sqr_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION exp FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class exp_op: public unary_op<T> {
    public:
	exp_op(const exp_op&);
	exp_op(const op<T>& a);
	virtual ~exp_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION log FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class LOGop: public unary_op<T> {
    public:
	LOGop(const LOGop&);
	LOGop(const op<T>& a);
	virtual ~LOGop();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION sqrt FORMULA                    */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class sqrt_op: public unary_op<T> {
    public:
	sqrt_op(const sqrt_op&);
	sqrt_op(const op<T>& a);
	virtual ~sqrt_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION sin FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class sin_op: public unary_op<T> {
    public:
	sin_op(const sin_op&);
	sin_op(const op<T>& a);
	virtual ~sin_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION cos FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class COSop: public unary_op<T> {
    public:
	COSop(const COSop&);
	COSop(const op<T>& a);
	virtual ~COSop();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION tan FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class TANop: public unary_op<T> {
    public:
	TANop(const TANop&);
	TANop(const op<T>& a);
	virtual ~TANop();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION asin FORMULA                    */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class Asin_op: public unary_op<T> {
    public:
	Asin_op(const Asin_op&);
	Asin_op(const op<T>& a);
	virtual ~Asin_op();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION acos FORMULA                    */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class ACOSop: public unary_op<T> {
    public:
	ACOSop(const ACOSop&);
	ACOSop(const op<T>& a);
	virtual ~ACOSop();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION atan FORMULA                    */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
class ATANop: public unary_op<T> {
    public:
	ATANop(const ATANop&);
	ATANop(const op<T>& a);
	virtual ~ATANop();
	virtual op<T>* copy() const;
	virtual void propagateop();
};
// ---------------------------------------------------------------------
// ICI
// ---------------------------------------------------------------------

/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION ADDITION FORMULAS               */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* --------------------------- op + op --------------------------------- */

template <class T>
inline add_op<T>::add_op(const add_op<T>& o):
	binary_op<T>(o)
{}

template <class T>
inline add_op<T>::add_op(const op<T>& a, const op<T>& b):
	binary_op<T>(a.m_val->value()+b.m_val->value(),a.copy(),b.copy())
{}

template <class T>
inline add_op<T>::~add_op()
{}

template <class T>
inline op<T>* add_op<T>::copy() const
{
	return new_copy_macro(add_op<T>,*this);
}

template <class T>
inline void add_op<T>::propagateop()
{
	this->m_o1->m_val->pv(*this->m_val);
	this->m_o2->m_val->pv(*this->m_val);
}
/* --------------------------------------------------------------------- */
/*                                                                       */
/*            BACKWARD DIFFERENTIATION SUBTRACTION FORMULAS              */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* --------------------------- op - op --------------------------------- */

template <class T>
inline sub_op<T>::sub_op(const sub_op& o):
	binary_op<T>(o)
{}

template <class T>
inline sub_op<T>::sub_op(const op<T>& a, const op<T>& b):
	binary_op<T>(a.m_val->value()-b.m_val->value(),a.copy(),b.copy())
{}

template <class T>
inline sub_op<T>::~sub_op()
{}

template <class T>
inline op<T>* sub_op<T>::copy() const
{
	return new_copy_macro(sub_op<T>,*this);
}

template <class T>
inline void sub_op<T>::propagateop()
{
	this->m_o1->m_val->pv(*this->m_val);
	this->m_o2->m_val->mv(*this->m_val);
}
/* --------------------------------------------------------------------- */
/*                                                                       */
/*          BACKWARD DIFFERENTIATION MULTIPLICATION FORMULAS             */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* --------------------------- op * op --------------------------------- */

template <class T>
inline mult_op<T>::mult_op(const mult_op& o):
	binary_op<T>(o)
{}

template <class T>
inline mult_op<T>::mult_op(const op<T>& a, const op<T>& b):
	binary_op<T>(a.m_val->value()*b.m_val->value(),a.copy(),b.copy())
{}

template <class T>
inline mult_op<T>::~mult_op()
{}

template <class T>
inline op<T>* mult_op<T>::copy() const
{
	return new_copy_macro(mult_op<T>,*this);
}

template <class T>
inline void mult_op<T>::propagateop()
{
	this->m_o1->m_val->pav(this->m_o2->m_val->value(),*this->m_val);
	this->m_o2->m_val->pav(this->m_o1->m_val->value(),*this->m_val);
}
/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION DIVISION FORMULAS               */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* --------------------------- op / op --------------------------------- */

template <class T>
inline div_op<T>::div_op(const div_op& o):
	binary_op<T>(o)
{}

template <class T>
inline div_op<T>::div_op(const op<T>& a, const op<T>& b):
	binary_op<T>(a.m_val->value()/b.m_val->value(),a.copy(),b.copy())
{}

template <class T>
inline div_op<T>::~div_op()
{}

template <class T>
inline op<T>* div_op<T>::copy() const
{
	return new_copy_macro(div_op<T>,*this);
}

template <class T>
inline void div_op<T>::propagateop()
{
	T tmp=1.0/this->m_o2->m_val->value();
	this->m_o1->m_val->pav(tmp,*this->m_val);
	this->m_o2->m_val->mav(tmp*this->m_val->value(),*this->m_val);
}
/* --------------------------------------------------------------------- */
/*                                                                       */
/*                BACKWARD DIFFERENTIATION POWER FORMULAS                */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* -------------------------- pow(op,op) ------------------------------- */

template <class T>
inline pow_op<T>::pow_op(const pow_op& o):
	binary_op<T>(o)
{}

template <class T>
inline pow_op<T>::pow_op(const op<T>& a, const op<T>& b):
	binary_op<T>(pow(a.m_val->value(),b.m_val->value()),a.copy(),b.copy())
{}

template <class T>
inline pow_op<T>::~pow_op()
{}

template <class T>
inline op<T>* pow_op<T>::copy() const
{
	return new_copy_macro(pow_op<T>,*this);
}

template <class T>
inline void pow_op<T>::propagateop()
{
	T tmp1(this->m_o2->m_val->value() *  pow(this->m_o1->m_val->value(),this->m_o2->m_val->value()-1.0));
	T tmp2(this->m_val->value() * log(this->m_o1->m_val->value()));
	this->m_o1->m_val->pav(tmp1,*this->m_val);
	this->m_o2->m_val->pav(tmp2,*this->m_val);
}
/* --------------------------------------------------------------------- */
/*                                                                       */
/*            BACKWARD DIFFERENTIATION UNARY + FORMULA                   */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline uadd_op<T>::uadd_op(const uadd_op& o):
	unary_op<T>(o)
{}

template <class T>
inline uadd_op<T>::uadd_op(const op<T>& a):
	unary_op<T>(a.m_val->value(),a.copy())
{}

template <class T>
inline uadd_op<T>::~uadd_op()
{}

template <class T>
inline op<T>* uadd_op<T>::copy() const
{
	return new_copy_macro(uadd_op<T>,*this);
}

template <class T>
inline void uadd_op<T>::propagateop()
{
	this->m_o1->m_val->pv(*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*            BACKWARD DIFFERENTIATION UNARY - FORMULA                   */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline usub_op<T>::usub_op(const usub_op& o):
	unary_op<T>(o)
{}

template <class T>
inline usub_op<T>::usub_op(const op<T>& a):
	unary_op<T>(-a.m_val->value(),a.copy())
{}

template <class T>
inline usub_op<T>::~usub_op()
{}

template <class T>
inline op<T>* usub_op<T>::copy() const
{
	return new_copy_macro(usub_op<T>,*this);
}

template <class T>
inline void usub_op<T>::propagateop()
{
	this->m_o1->m_val->mv(*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*          BACKWARD DIFFERENTIATION pow(op,int) FORMULA                 */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline pow_n_op<T>::pow_n_op(const pow_n_op& o):
	unary_op<T>(o),m_b(o.m_b)
{}

template <class T>
inline pow_n_op<T>::pow_n_op(const op<T>& a, int ib):
	unary_op<T>(pow(a.m_val->value(),ib),a.copy()),m_b(ib)
{}

template <class T>
inline pow_n_op<T>::~pow_n_op()
{}

template <class T>
inline op<T>* pow_n_op<T>::copy() const
{
	return new_copy_macro(pow_n_op<T>,*this);
}

template <class T>
inline void pow_n_op<T>::propagateop()
{
	T tmp(m_b*pow(this->m_o1->m_val->value(),m_b-1));
	this->m_o1->m_val->pav(tmp,*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION sqr FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline sqr_op<T>::sqr_op(const sqr_op& o):
	unary_op<T>(o)
{}

template <class T>
inline sqr_op<T>::sqr_op(const op<T>& a):
	unary_op<T>(_sqr(a.m_val->value()),a.copy())
{}

template <class T>
inline sqr_op<T>::~sqr_op()
{}

template <class T>
inline op<T>* sqr_op<T>::copy() const
{
	return new_copy_macro(sqr_op<T>,*this);
}

template <class T>
inline void sqr_op<T>::propagateop()
{
	T tmp(2.0*(this->m_o1->m_val)->value());
	this->m_o1->m_val->pav(tmp,*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION exp FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline exp_op<T>::exp_op(const exp_op& o):
	unary_op<T>(o)
{}

template <class T>
inline exp_op<T>::exp_op(const op<T>& a):
	unary_op<T>(exp(a.m_val->value()),a.copy())
{}

template <class T>
inline exp_op<T>::~exp_op()
{}

template <class T>
op<T>* exp_op<T>::copy() const
{
	return new_copy_macro(exp_op<T>,*this);
}

template <class T>
void exp_op<T>::propagateop()
{
	this->m_o1->m_val->pav(this->m_val->value(),*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION log FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline LOGop<T>::LOGop(const LOGop& o):
	unary_op<T>(o)
{}

template <class T>
inline LOGop<T>::LOGop(const op<T>& a):
	unary_op<T>(log(a.m_val->value()),a.copy())
{}

template <class T>
inline LOGop<T>::~LOGop()
{}

template <class T>
op<T>* LOGop<T>::copy() const
{
	return new_copy_macro(LOGop<T>,*this);
}

template <class T>
void LOGop<T>::propagateop()
{
	this->m_o1->m_val->pav(1.0/this->m_o1->m_val->value(),*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION sqrt FORMULA                    */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline sqrt_op<T>::sqrt_op(const sqrt_op& o):
	unary_op<T>(o)
{}

template <class T>
inline sqrt_op<T>::sqrt_op(const op<T>& a):
	unary_op<T>(sqrt(a.m_val->value()),a.copy())
{}

template <class T>
inline sqrt_op<T>::~sqrt_op()
{}

template <class T>
inline op<T>* sqrt_op<T>::copy() const
{
	return new_copy_macro(sqrt_op<T>,*this);
}

template <class T>
inline void sqrt_op<T>::propagateop()
{
	T tmp(1.0/(this->m_val->value()*2.0));
	this->m_o1->m_val->pav(tmp,*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION sin FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline sin_op<T>::sin_op(const sin_op& o):
	unary_op<T>(o)
{}

template <class T>
inline sin_op<T>::sin_op(const op<T>& a):
	unary_op<T>(sin(a.m_val->value()),a.copy())
{}

template <class T>
inline sin_op<T>::~sin_op()
{}

template <class T>
inline op<T>* sin_op<T>::copy() const
{
	return new_copy_macro(sin_op<T>,*this);
}

template <class T>
inline void sin_op<T>::propagateop()
{
	T tmp(cos((this->m_o1->m_val)->value()));
	this->m_o1->m_val->pav(tmp,*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION cos FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline COSop<T>::COSop(const COSop& o):
	unary_op<T>(o)
{}

template <class T>
inline COSop<T>::COSop(const op<T>& a):
	unary_op<T>(cos(a.m_val->value()),a.copy())
{}

template <class T>
inline COSop<T>::~COSop()
{}

template <class T>
inline op<T>* COSop<T>::copy() const
{
	return new_copy_macro(COSop<T>,*this);
}

template <class T>
inline void COSop<T>::propagateop()
{
	T tmp(sin((this->m_o1->m_val)->value()));
	this->m_o1->m_val->mav(tmp,*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION tan FORMULA                     */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline TANop<T>::TANop(const TANop& o):
	unary_op<T>(o)
{}

template <class T>
inline TANop<T>::TANop(const op<T>& a):
	unary_op<T>(tan(a.m_val->value()),a.copy())
{}

template <class T>
inline TANop<T>::~TANop()
{}

template <class T>
inline op<T>* TANop<T>::copy() const
{
	return new_tab_macro(TANop<T>,*this);
}

template <class T>
inline void TANop<T>::propagateop()
{
	T tmp(_sqr(this->m_val->value())+1.0);
	this->m_o1->m_val->pav(tmp,*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION asin FORMULA                    */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline Asin_op<T>::Asin_op(const Asin_op& o):
	unary_op<T>(o)
{}

template <class T>
inline Asin_op<T>::Asin_op(const op<T>& a):
	unary_op<T>(asin(a.m_val->value()),a.copy())
{}

template <class T>
inline Asin_op<T>::~Asin_op()
{}

template <class T>
inline op<T>* Asin_op<T>::copy() const
{
	return new_tab_macro(Asin_op<T>,*this);
}

template <class T>
inline void Asin_op<T>::propagateop()
{
	T tmp(1.0/sqrt(1.0-_sqr(this->m_o1->m_val->value())));
	this->m_o1->m_val->pav(tmp,*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION acos FORMULA                    */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline ACOSop<T>::ACOSop(const ACOSop& o):
	unary_op<T>(o)
{}

template <class T>
inline ACOSop<T>::ACOSop(const op<T>& a):
	unary_op<T>(acos(a.m_val->value()),a.copy())
{}

template <class T>
inline ACOSop<T>::~ACOSop()
{}

template <class T>
inline op<T>* ACOSop<T>::copy() const
{
	return new_tab_macro(ACOSop<T>,*this);
}

template <class T>
inline void ACOSop<T>::propagateop()
{
	T tmp(1.0/sqrt(1.0-_sqr(this->m_o1->m_val->value())));
	this->m_o1->m_val->mav(tmp,*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*              BACKWARD DIFFERENTIATION atan FORMULA                    */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
ATANop<T>::ATANop(const ATANop& o):
	unary_op<T>(o)
{}

template <class T>
inline ATANop<T>::ATANop(const op<T>& a):
	unary_op<T>(atan(a.m_val->value()),a.copy())
{}

template <class T>
inline ATANop<T>::~ATANop()
{}

template <class T>
inline op<T>* ATANop<T>::copy() const
{
	return new_tab_macro(ATANop<T>,*this);
}

template <class T>
inline void ATANop<T>::propagateop()
{
	T tmp(1.0/(_sqr(this->m_o1->m_val->value())+1.0));
	this->m_o1->m_val->pav(tmp,*this->m_val);
}

/* --------------------------------------------------------------------- */
/*                                                                       */
/*                        DEFINITION OF OPERATORS                        */
/*                                                                       */
/* --------------------------------------------------------------------- */

template <class T>
inline add_op<T> operator+ (const op<T>& a, const op<T>& b)
{
	return add_op<T>(a,b);
}

template <class T>
inline add_op<T> operator+ (const op<T>& a, const dag<T>& b)
{
	return add_op<T>(a,b);
}

template <class T>
inline add_op<T> operator+ (const dag<T>& a, const op<T>& b)
{
	return add_op<T>(a,b);
}

template <class T>
inline add_op<T> operator+ (const dag<T>& a, const dag<T>& b)
{
	return add_op<T>(a,b);
}
template <class T>
inline sub_op<T> operator- (const op<T>& a, const op<T>& b)
{
	return sub_op<T>(a,b);
}

template <class T>
inline sub_op<T> operator- (const op<T>& a, const dag<T>& b)
{
	return sub_op<T>(a,b);
}

template <class T>
inline sub_op<T> operator- (const dag<T>& a, const op<T>& b)
{
	return sub_op<T>(a,b);
}

template <class T>
inline sub_op<T> operator- (const dag<T>& a, const dag<T>& b)
{
	return sub_op<T>(a,b);
}

template <class T>
inline mult_op<T> operator* (const op<T>& a, const op<T>& b)
{
	return mult_op<T>(a,b);
}

template <class T>
inline mult_op<T> operator* (const op<T>& a, const dag<T>& b)
{
	return mult_op<T>(a,b);
}

template <class T>
inline mult_op<T> operator* (const dag<T>& a, const op<T>& b)
{
	return mult_op<T>(a,b);
}

template <class T>
inline mult_op<T> operator* (const dag<T>& a, const dag<T>& b)
{
	return mult_op<T>(a,b);
}

template <class T>
inline div_op<T> operator/ (const op<T>& a, const op<T>& b)
{
	return div_op<T>(a,b);
}

template <class T>
inline div_op<T> operator/ (const op<T>& a, const dag<T>& b)
{
	return div_op<T>(a,b);
}

template <class T>
inline div_op<T> operator/ (const dag<T>& a, const op<T>& b)
{
	return div_op<T>(a,b);
}

template <class T>
inline div_op<T> operator/ (const dag<T>& a, const dag<T>& b)
{
	return div_op<T>(a,b);
}


template <class T>
inline pow_op<T> pow (const op<T>& a, const op<T>& b)
{
	return pow_op<T>(a,b);
}

template <class T>
inline pow_op<T> pow (const op<T>& a, const dag<T>& b)
{
	return pow_op<T>(a,b);
}

template <class T>
inline pow_op<T> pow (const dag<T>& a, const op<T>& b)
{
	return pow_op<T>(a,b);
}

template <class T>
inline pow_op<T> pow (const dag<T>& a, const dag<T>& b)
{
	return pow_op<T>(a,b);
}

template <class T>
inline uadd_op<T> operator+ (const op<T>& a)
{
	return uadd_op<T>(a);
}

template <class T>
inline usub_op<T> operator- (const op<T>& a)
{
	return usub_op<T>(a);
}

template <class T>
inline pow_n_op<T> pow (const op<T>& a, int ib)
{
	return pow_n_op<T>(a,ib);
}

template <class T>
inline pow_n_op<T> pow (const dag<T>& a, int ib)
{
	return pow_n_op<T>(a,ib);
}

template <class T>
inline sqr_op<T> sqr (const op<T>& a)
{
	return sqr_op<T>(a);
}

template <class T>
inline exp_op<T> exp (const op<T>& a)
{
	return exp_op<T>(a);
}

template <class T>
inline LOGop<T> log (const op<T>& a)
{
	return LOGop<T>(a);
}

template <class T>
inline sqrt_op<T> sqrt (const op<T>& a)
{
	return sqrt_op<T>(a);
}

template <class T>
inline sin_op<T> sin (const op<T>& a)
{
	return sin_op<T>(a);
}

template <class T>
inline COSop<T> cos (const op<T>& a)
{
	return COSop<T>(a);
}

template <class T>
inline TANop<T> tan (const op<T>& a)
{
	return TANop<T>(a);
}

template <class T>
inline Asin_op<T> asin (const op<T>& a)
{
	return Asin_op<T>(a);
}

template <class T>
inline ACOSop<T> acos (const op<T>& a)
{
	return ACOSop<T>(a);
}

template <class T>
inline ATANop<T> atan (const op<T>& a)
{
	return ATANop<T>(a);
}

template <class T>
inline dag<T>& dag<T>::operator += (const op<T>& x)
{
	return *this = (*this)+x;
}

template <class T>
inline dag<T>& dag<T>::operator -= (const op<T>& x)
{
	return *this = (*this)-x;
}

template <class T>
inline dag<T>& dag<T>::operator *= (const op<T>& x)
{
	return *this = (*this)*x;
}

template <class T>
inline dag<T>& dag<T>::operator /= (const op<T>& x)
{
	return *this = (*this)/x;
}

template <class T>
inline dag<T>& dag<T>::operator += (const dag<T>& x)
{
	return *this = (*this)+x;
}

template <class T>
inline dag<T>& dag<T>::operator -= (const dag<T>& x)
{
	return *this = (*this)-x;
}

template <class T>
inline dag<T>& dag<T>::operator *= (const dag<T>& x)
{
	return *this = (*this)*x;
}

template <class T>
inline dag<T>& dag<T>::operator /= (const dag<T>& x)
{
	return *this = (*this)/x;
}

template <class T>
inline dag<T>& dag<T>::operator += (const T& x)
{
	return *this = (*this)+x;
}

template <class T>
inline dag<T>& dag<T>::operator -= (const T& x)
{
	return *this = (*this)-x;
}

template <class T>
inline dag<T>& dag<T>::operator *= (const T& x)
{
	return *this = (*this)*x;
}

template <class T>
inline dag<T>& dag<T>::operator /= (const T& x)
{
	return *this = (*this)/x;
}
// -----------------------------------------------------------------------------------
#ifdef HASEQ
template <class T>
inline bool operator == (const op<T> &a, const op<T> &b)
{
	basic_assert_macro(a.m_val && b.m_val);
	return a.m_val->value()==b.m_val->value();
};
#endif

#ifdef HASNEQ
template <class T>
inline bool operator != (const op<T> &a, const op<T> &b)
{
	basic_assert_macro(a.m_val && b.m_val);
	return (a.m_val->value()!=b.m_val->value());
}
#endif

#ifdef HASGT
template <class T>
inline bool operator > (const op<T> &a, const op<T> &b)
{
	basic_assert_macro(a.m_val && b.m_val);
	return (a.m_val->value()>b.m_val->value());
}
#endif

#ifdef HASGEQ
template <class T>
inline bool operator >= (const op<T> &a, const op<T> &b)
{
	basic_assert_macro(a.m_val && b.m_val);
	return (a.m_val->value()>=b.m_val->value());
}
#endif

#ifdef HASLT
template <class T>
inline bool operator < (const op<T> &a, const op<T> &b)
{
	basic_assert_macro(a.m_val && b.m_val);
	return (a.m_val->value()<b.m_val->value());
}
#endif

#ifdef HASLEQ
template <class T>
inline bool operator <= (const op<T> &a, const op<T> &b)
{
	basic_assert_macro(a.m_val && b.m_val);
	return (a.m_val->value()<=b.m_val->value());
}
#endif
